package com.sunyard.chsm.service.impl;

import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.conditions.update.LambdaUpdateWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.sunyard.chsm.dto.KeyInfoDTO;
import com.sunyard.chsm.enums.KeyCategory;
import com.sunyard.chsm.enums.KeyStatus;
import com.sunyard.chsm.enums.KeyUsage;
import com.sunyard.chsm.mapper.KeyInfoMapper;
import com.sunyard.chsm.mapper.KeyTemplateMapper;
import com.sunyard.chsm.mapper.SpKeyRecordMapper;
import com.sunyard.chsm.model.entity.KeyInfo;
import com.sunyard.chsm.model.entity.KeyRecord;
import com.sunyard.chsm.model.entity.KeyTemplate;
import com.sunyard.chsm.sdf.SdfApiService;
import com.sunyard.chsm.sdf.model.EccKey;
import com.sunyard.chsm.service.KeyInfoService;
import com.sunyard.chsm.utils.JsonUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Hex;
import org.springframework.beans.BeanUtils;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;
import org.springframework.util.CollectionUtils;
import org.springframework.util.StringUtils;

import javax.annotation.Resource;
import java.nio.charset.StandardCharsets;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * @author liulu
 * @since 2024/10/23
 */
@Slf4j
@Service
public class KeyInfoServiceImpl implements KeyInfoService {

    @Resource
    private KeyInfoMapper keyInfoMapper;
    @Resource
    private SpKeyRecordMapper spKeyRecordMapper;
    @Resource
    private KeyTemplateMapper keyTemplateMapper;
    @Resource
    private SdfApiService sdfApiService;


    @Override
    public Page<KeyInfoDTO.View> selectPageList(KeyInfoDTO.Query query) {

        IPage<KeyInfo> page = keyInfoMapper.selectPage(
                new Page<>(query.getPageNumber(), query.getPageSize()),
                new LambdaQueryWrapper<KeyInfo>()
                        .eq(StringUtils.hasText(query.getKeyType()), KeyInfo::getKeyType, query.getKeyType())
                        .eq(StringUtils.hasText(query.getStatus()), KeyInfo::getStatus, query.getStatus())
                        .orderByDesc(KeyInfo::getCreateTime)
        );
        List<KeyInfo> records = page.getRecords();
        if (CollectionUtils.isEmpty(records)) {
            return new Page<>(page.getCurrent(), page.getSize(), page.getTotal());
        }
        List<KeyInfoDTO.View> viewList = records.stream()
                .map(it -> {
                    KeyInfoDTO.View view = new KeyInfoDTO.View();
                    BeanUtils.copyProperties(it, view);
                    Optional.ofNullable(KeyCategory.of(it.getKeyType()))
                            .map(KeyCategory::getDesc)
                            .ifPresent(view::setKeyTypeText);
                    Map<String, String> usageMap = KeyUsage.getUsage(it.getKeyUsage())
                            .stream()
                            .collect(Collectors.toMap(KeyUsage::getCode, KeyUsage::getDesc));
                    view.setKeyUsages(new ArrayList<>(usageMap.keySet()));
                    view.setKeyUsageText(String.join(",", usageMap.values()));
                    Optional.ofNullable(KeyStatus.of(it.getStatus()))
                            .map(KeyStatus::getDesc)
                            .ifPresent(view::setStatusText);

                    return view;
                })
                .collect(Collectors.toList());

        return new Page<KeyInfoDTO.View>(page.getCurrent(), page.getSize(), page.getTotal()).setRecords(viewList);
    }

    @Transactional
    @Override
    public Long save(KeyInfoDTO.Save save) {

        KeyTemplate keyTemplate = keyTemplateMapper.selectOne(
                new LambdaQueryWrapper<KeyTemplate>()
                        .eq(KeyTemplate::getCode, save.getKeyTemplateCode())
        );
        Assert.notNull(keyTemplate, "密钥模版不存在");

        LocalDateTime now = LocalDateTime.now();

        for (int i = 0; i < save.getGenNumber(); i++) {

            // 密钥信息
            KeyInfo info = new KeyInfo();
            info.setId(IdWorker.getId());
            info.setApplicationId(save.getApplicationId());
            info.setKeyTemplateId(keyTemplate.getId());
            info.setCode(String.valueOf(info.getId()));
            info.setKeyType(keyTemplate.getKeyType());
            info.setKeyAlg(keyTemplate.getKeyAlg());
            info.setKeyLength(keyTemplate.getKeyLength());
            info.setKeyUsage(keyTemplate.getKeyUsage());
            info.setCheckAlg(keyTemplate.getCheckAlg());
            info.setCreateTime(LocalDateTime.now());

            info.setStatus(KeyStatus.ENABLED.getCode());

            ChronoUnit startUnit = ChronoUnit.valueOf(keyTemplate.getStartAfterCreateUnit().toUpperCase());
            info.setEffectiveTime(now.plusSeconds(startUnit.getDuration().getSeconds() * keyTemplate.getStartAfterCreateTime()));

            ChronoUnit validUnit = ChronoUnit.valueOf(keyTemplate.getValidUnit().toUpperCase());
            info.setExpiredTime(info.getEffectiveTime().plusSeconds(validUnit.getDuration().getSeconds() * keyTemplate.getValidTime()));

            // 具体使用的密钥值
            KeyRecord record = genKeyRecord(info);
            info.setCheckValue(record.getCheckValue());
            keyInfoMapper.insert(info);
            spKeyRecordMapper.insert(record);
        }

        return 0L;
    }

    @Transactional
    @Override
    public void update(KeyInfoDTO.Update update) {
        List<Long> ids = update.getIds();
        if (CollectionUtils.isEmpty(ids)) {
            return;
        }
        LocalDateTime now = LocalDateTime.now();
        LocalDateTime newEffectTime = LocalDateTime.of(update.getEffectiveTime(), LocalTime.MIN);

        Assert.isTrue(now.isBefore(newEffectTime), "生效日期必须在当前时间之后");

        List<KeyInfo> keyInfos = keyInfoMapper.selectBatchIds(ids);
        if (CollectionUtils.isEmpty(keyInfos)) {
            log.warn("enableKey no exist key with ids: {}", ids.stream().map(String::valueOf).collect(Collectors.joining(",")));
            return;
        }
        List<String> unNormalCodes = keyInfos.stream()
                .filter(it -> now.isBefore(it.getEffectiveTime())
                        || now.isAfter(it.getExpiredTime())
                        || newEffectTime.isAfter(it.getExpiredTime())
                        || !Objects.equals(KeyStatus.ENABLED.getCode(), it.getStatus())
                )
                .map(KeyInfo::getCode)
                .collect(Collectors.toList());
        Assert.isTrue(CollectionUtils.isEmpty(unNormalCodes),
                "密钥id: " + String.join(",", unNormalCodes) + "不是启用状态或者新生效时间超过密钥过期时间, 无法更新");

        // 更新实际密钥值
        spKeyRecordMapper.update(
                new LambdaUpdateWrapper<KeyRecord>()
                        .set(KeyRecord::getExpiredTime, newEffectTime)
                        .in(KeyRecord::getKeyId, ids)
                        .lt(KeyRecord::getEffectiveTime, now)
                        .gt(KeyRecord::getExpiredTime, now)
        );

        for (KeyInfo info : keyInfos) {
            // 具体使用的密钥值
            KeyRecord record = genKeyRecord(info);

            KeyInfo upInfo = new KeyInfo();
            upInfo.setId(info.getId());
            upInfo.setCheckValue(record.getCheckValue());

            keyInfoMapper.updateById(upInfo);
            spKeyRecordMapper.insert(record);
        }

    }

    @Override
    public byte[] backupKey(KeyInfoDTO.Backup backup) {

        LocalDate plussed = backup.getStartTime().plusDays(100);
        Assert.isTrue(plussed.isAfter(backup.getEndTime()), "备份密钥创建时间范围不能超过100天");
        List<KeyInfo> keyInfos = keyInfoMapper.selectList(
                new LambdaQueryWrapper<KeyInfo>()
                        .eq(KeyInfo::getKeyType, backup.getKeyType())
                        .gt(KeyInfo::getCreateTime, LocalDateTime.of(backup.getStartTime(), LocalTime.MIN))
                        .lt(KeyInfo::getCreateTime, LocalDateTime.of(backup.getEndTime(), LocalTime.MAX))
        );

        StringBuilder builder = new StringBuilder();
        for (KeyInfo keyInfo : keyInfos) {
            List<KeyRecord> records = spKeyRecordMapper.selectList(
                    new LambdaQueryWrapper<KeyRecord>()
                            .eq(KeyRecord::getKeyId, keyInfo.getId())
            );
            keyInfo.setRecords(records);
            builder.append(JsonUtils.toJsonString(records)).append(System.lineSeparator());
        }
        return builder.toString().getBytes(StandardCharsets.UTF_8);
    }

    private KeyRecord genKeyRecord(KeyInfo info) {
        KeyRecord record = new KeyRecord();
        record.setId(IdWorker.getId());
        record.setKeyId(info.getId());
        record.setKeyIndex(String.valueOf(record.getId()));
        record.setCheckAlg(info.getCheckAlg());
        record.setEffectiveTime(info.getEffectiveTime());
        record.setExpiredTime(info.getExpiredTime());
        record.setCreateTime(LocalDateTime.now());

        if (KeyCategory.SYM_KEY.getCode().equals(info.getKeyType())) {
            byte[] symKey = sdfApiService.generateRandom(16);
            record.setKeyData(Hex.encodeHexString(symKey));
            String checkHash = Hex.encodeHexString(sdfApiService.hash(symKey));
            record.setCheckValue(checkHash);
        } else {
            EccKey eccKey = sdfApiService.genKeyPairEcc();
            byte[] d = eccKey.getPriKey().getD();
            record.setKeyData(Hex.encodeHexString(d));
            String checkHash = Hex.encodeHexString(sdfApiService.hash(d));
            record.setCheckValue(checkHash);

            byte[] pubKeyBytes = eccKey.getPubKey().getPubKeyBytes();
            record.setPubKey(Hex.encodeHexString(pubKeyBytes));
        }
        return record;
    }

    @Override
    public void enableKey(List<Long> ids) {
        if (CollectionUtils.isEmpty(ids)) {
            return;
        }
        List<KeyInfo> keyInfos = keyInfoMapper.selectBatchIds(ids);
        if (CollectionUtils.isEmpty(keyInfos)) {
            log.warn("enableKey no exist key with ids: {}", ids.stream().map(String::valueOf).collect(Collectors.joining(",")));
            return;
        }
        List<String> unNormalCodes = keyInfos.stream()
                .filter(it -> !Objects.equals(KeyStatus.DISABLED.getCode(), it.getStatus()))
                .map(KeyInfo::getCode)
                .collect(Collectors.toList());
        Assert.isTrue(CollectionUtils.isEmpty(unNormalCodes),
                "密钥id: " + String.join(",", unNormalCodes) + "不是已停用状态, 无法启用");
        keyInfoMapper.update(
                new LambdaUpdateWrapper<KeyInfo>()
                        .set(KeyInfo::getStatus, KeyStatus.ENABLED.getCode())
                        .in(KeyInfo::getId, ids)
        );
    }

    @Override
    public void disableKey(List<Long> ids) {
        if (CollectionUtils.isEmpty(ids)) {
            return;
        }
        List<KeyInfo> keyInfos = keyInfoMapper.selectBatchIds(ids);
        if (CollectionUtils.isEmpty(keyInfos)) {
            log.warn("disableKey no exist key with ids: {}", ids.stream().map(String::valueOf).collect(Collectors.joining(",")));
            return;
        }
        List<String> unNormalCodes = keyInfos.stream()
                .filter(it -> !Objects.equals(KeyStatus.ENABLED.getCode(), it.getStatus()))
                .map(KeyInfo::getCode)
                .collect(Collectors.toList());
        Assert.isTrue(CollectionUtils.isEmpty(unNormalCodes),
                "密钥id: " + String.join(",", unNormalCodes) + "不是启用状态, 无法停用");
        keyInfoMapper.update(
                new LambdaUpdateWrapper<KeyInfo>()
                        .set(KeyInfo::getStatus, KeyStatus.DISABLED.getCode())
                        .in(KeyInfo::getId, ids)
        );
    }

    @Override
    public void archiveKey(List<Long> ids) {

        if (CollectionUtils.isEmpty(ids)) {
            return;
        }
        List<KeyInfo> keyInfos = keyInfoMapper.selectBatchIds(ids);
        if (CollectionUtils.isEmpty(keyInfos)) {
            log.warn("archiveKey no exist key with ids: {}", ids.stream().map(String::valueOf).collect(Collectors.joining(",")));
            return;
        }
//        List<String> unNormalCodes = keyInfos.stream()
//                .filter(it -> !Objects.equals(KeyStatus.ENABLED.getCode(), it.getStatus()))
//                .map(KeyInfo::getCode)
//                .collect(Collectors.toList());
//        Assert.isTrue(CollectionUtils.isEmpty(unNormalCodes),
//                "密钥id: " + String.join(",", unNormalCodes) + "不是启用状态, 无法停用");
        keyInfoMapper.update(
                new LambdaUpdateWrapper<KeyInfo>()
                        .set(KeyInfo::getStatus, KeyStatus.ARCHIVED.getCode())
                        .in(KeyInfo::getId, ids)
        );
    }

    @Override
    public void recoveryKey(List<Long> ids) {

        if (CollectionUtils.isEmpty(ids)) {
            return;
        }
        List<KeyInfo> keyInfos = keyInfoMapper.selectBatchIds(ids);
        if (CollectionUtils.isEmpty(keyInfos)) {
            log.warn("recoveryKey no exist key with ids: {}", ids.stream().map(String::valueOf).collect(Collectors.joining(",")));
            return;
        }
        List<String> unNormalCodes = keyInfos.stream()
                .filter(it -> !Objects.equals(KeyStatus.ARCHIVED.getCode(), it.getStatus()))
                .map(KeyInfo::getCode)
                .collect(Collectors.toList());
        Assert.isTrue(CollectionUtils.isEmpty(unNormalCodes),
                "密钥id: " + String.join(",", unNormalCodes) + "不是归档状态, 无法恢复");
        keyInfoMapper.update(
                new LambdaUpdateWrapper<KeyInfo>()
                        .set(KeyInfo::getStatus, KeyStatus.DISABLED.getCode())
                        .in(KeyInfo::getId, ids)
        );
    }

    @Override
    public void destroyKey(List<Long> ids) {
        if (CollectionUtils.isEmpty(ids)) {
            return;
        }
        List<KeyInfo> keyInfos = keyInfoMapper.selectBatchIds(ids);
        if (CollectionUtils.isEmpty(keyInfos)) {
            log.warn("destroyKey no exist key with ids: {}", ids.stream().map(String::valueOf).collect(Collectors.joining(",")));
            return;
        }
        List<String> unNormalCodes = keyInfos.stream()
                .filter(it -> !Objects.equals(KeyStatus.ARCHIVED.getCode(), it.getStatus()))
                .map(KeyInfo::getCode)
                .collect(Collectors.toList());
        Assert.isTrue(CollectionUtils.isEmpty(unNormalCodes),
                "密钥id: " + String.join(",", unNormalCodes) + "不是归档状态, 无法销毁");
        keyInfoMapper.update(
                new LambdaUpdateWrapper<KeyInfo>()
                        .set(KeyInfo::getStatus, KeyStatus.DESTORY.getCode())
                        .in(KeyInfo::getId, ids)
        );
    }
}
